﻿using System;
using System.Collections.Generic;
using System.Linq;
using DependencyAnalysis.DataModel;

namespace DependencyAnalysis
{
    public abstract class Metrics
    {
        private readonly HalsteadCache _halsteadCache = new HalsteadCache();

        private readonly IDependencyModel _model;

        protected Metrics(IDependencyModel model)
        {
            _model = model;
        }

        /// <summary>
        /// With the number of referrers, then the impact of changing API and implementation grows. 
        /// Return the number of types in other modules than the target that refer to the target module.
        /// </summary>
        /// <param name="target">a module</param>
        /// <returns>number of referred types</returns>
        public int AfferentCoupling(ModuleDescriptor target)
        {
            if (!_model.DependenciesPerModule.ContainsKey(target)) return 0;
            var types = VerifierUtilities.AggregateDependenciesToTypes(_model.DependenciesPerModule[target], true);
            return types.Count();
        }

        /// <summary>
        /// With the number of referrers, then the impact of changing API and implementation grows. 
        /// Return the number of types in other submodules than the target that refer to the target submodule.
        /// </summary>
        /// <param name="target">a submodule</param>
        /// <returns>number of referred types</returns>
        public int AfferentCoupling(SubmoduleDescriptor target)
        {
            if (!_model.DependenciesPerSubmodule.ContainsKey(target)) return 0;
            var types = VerifierUtilities.AggregateDependenciesToTypes(_model.DependenciesPerSubmodule[target], true);
            return types.Count();
        }

        /// <summary>
        /// With the number of referrers, then the impact of changing member signatures and implementation grows.
        /// Return the number of types that refer to the examined type. Self references are excluded.
        /// </summary>
        /// <param name="target">a type</param>
        /// <returns>number of referred types</returns>
        public int AfferentCoupling(TypeDescriptor target)
        {
            if (!_model.DependenciesPerType.ContainsKey(target)) return 0;
            var types = VerifierUtilities.AggregateDependenciesToTypes(_model.DependenciesPerType[target], true);
            return types.Count();
        }

        /// <summary>
        /// With the number of referrers, then the impact of changing the members signature and implementation grows.
        /// Return the number of members that refer to the examined member. Self references are excluded.
        /// </summary>
        /// <param name="target">a member</param>
        /// <returns>number of referred members</returns>
        public int AfferentCoupling(MemberDescriptor target)
        {
            if (!_model.DependenciesPerMember.ContainsKey(target)) return 0;
            return _model.DependenciesPerMember[target].Count;
        }

        /// <summary>
        /// Return the number of types the examined module depends on. The type inside the module are excluded.
        /// </summary>
        /// <param name="target">a module</param>
        /// <returns>number of referred types</returns>
        public int EfferentCoupling(ModuleDescriptor target)
        {
            if (!_model.DependenciesPerModuleOut.ContainsKey(target)) return 0;
            var types = VerifierUtilities.AggregateDependenciesToTypes(_model.DependenciesPerModuleOut[target], false);
            return types.Count();
        }

        /// <summary>
        /// Return the number of types the examined submodule depends on. The type inside the submodule are excluded.
        /// </summary>
        /// <param name="target">a submodule</param>
        /// <returns>number of referred types</returns>
        public int EfferentCoupling(SubmoduleDescriptor target)
        {
            if (!_model.DependenciesPerSubmoduleOut.ContainsKey(target)) return 0;
            var types = VerifierUtilities.AggregateDependenciesToTypes(_model.DependenciesPerSubmoduleOut[target], false);
            return types.Count();
        }

        /// <summary>
        /// Return the number of types the examined type depends on. Self references are excluded.
        /// </summary>
        /// <param name="target">a type</param>
        /// <returns>number of referred types</returns>
        public int EfferentCoupling(TypeDescriptor target)
        {
            if (!_model.DependenciesPerTypeOut.ContainsKey(target)) return 0;
            var types = VerifierUtilities.AggregateDependenciesToTypes(_model.DependenciesPerTypeOut[target], false);
            return types.Count();
        }

        /// <summary>
        /// Return the number of members the examined member depends on. The declaring type is excluded.
        /// </summary>
        /// <param name="target">a member</param>
        /// <returns>number of referred members</returns>
        public int EfferentCoupling(MemberDescriptor target)
        {
            if (!_model.DependenciesPerMemberOut.ContainsKey(target)) return 0;
            return _model.DependenciesPerMemberOut[target].Count;
        }

        /// <summary>
        /// The average number of relations between types of the module per type.
        /// Optimal values lie between 1.5 and 4.0.
        /// RelationalCohesion = (RelationShips + 1) / Types
        /// </summary>
        /// <param name="target">a module</param>
        /// <returns></returns>
        public double RelationalCohesion(ModuleDescriptor target)
        {
            if (!_model.DependenciesPerModule.ContainsKey(target)) return 0;
            int relationships = 0;
            var typesInModule = target.Submodules.SelectMany(space => space.Types).ToArray();
            if (typesInModule.Length == 0) return 0;
            foreach (var typeDescriptor in typesInModule)
            {
                var dependencies = _model.DependenciesPerType[typeDescriptor];
                var internalReferences = dependencies.Where(dependency => dependency.Source.ImplementingType.Submodule.Module.Equals(target)).ToArray();
                var types = VerifierUtilities.AggregateDependenciesToTypes(internalReferences, true);
                relationships += types.Count();
            }
            return (relationships + 1) / (double)typesInModule.Length;
        }

        /// <summary>
        /// Modules that contain multiple outgoing but few incoming dependencies are less stable because of the consequences of changes in these Modules.
        /// Modules containing more incoming dependencies are more stable because they are more difficult to change.
        /// Instability = EfferentCoupling / (EfferentCoupling + AfferentCoupling)
        /// </summary>
        /// <param name="target">a submodule</param>
        /// <returns>[0, 1]</returns>
        public double Instability(ModuleDescriptor target)
        {
            if (!_model.DependenciesPerModule.ContainsKey(target)) return 0;
            var efferent = EfferentCoupling(target);
            var afferent = AfferentCoupling(target);
            if (efferent + afferent == 0) return 1;
            return efferent / (double)(efferent + afferent);
        }

        /// <summary>
        /// The ratio of abstract and interface types in a module.
        /// </summary>
        /// <param name="target">a module</param>
        /// <returns>[0, 1]</returns>
        public double Abstractness(ModuleDescriptor target)
        {
            if (!_model.DependenciesPerModule.ContainsKey(target)) return 0;
            var typesInModule = target.Submodules.SelectMany(space => space.Types).ToArray();
            if (typesInModule.Length == 0) return 0;
            int abstractCount = typesInModule.Count(typeDescriptor =>
                typeDescriptor.AbstractionType == AbstractionType.Interface || typeDescriptor.AbstractionType == AbstractionType.Abstract);
            return abstractCount / (double)typesInModule.Length;
        }

        /// <summary>
        /// Describes the balance between abstractness and stability of a module. A module becomes stable when there are a lot of
        /// referrers to it. If so, then it should be abstract at the same time. A module that is instable should be concrete, as there
        /// are no referrers relying on contracts.
        /// Values are in range [-1, +1]. Close to -1 means the module is concrete but has many dependent modules, changes to the module will
        /// imply lots of changes to referrers. This is called the zone of pain. On the other side, a value of one means the module is
        /// almost completely abstract, but noone uses it. Which is the zone of uselessness.
        /// NormalizedDistance = Abstractness + Instability - 1
        /// </summary>
        /// <param name="target">a module</param>
        /// <returns>[-1, +1]</returns>
        public double DistanceFromMainSequence(ModuleDescriptor target)
        {
            var abstractness = Abstractness(target);
            var instability = Instability(target);
            return abstractness + instability - 1;
        }

        /// <summary>
        /// The average number of relations between types of the submodule per type.
        /// Optimal values lie between 1.5 and 4.0.
        /// RelationalCohesion = (RelationShips + 1) / Types
        /// </summary>
        /// <param name="target">a submodule</param>
        /// <returns></returns>
        public double RelationalCohesion(SubmoduleDescriptor target)
        {
            if (!_model.DependenciesPerSubmodule.ContainsKey(target)) return 0;
            var internalReferences = new List<DependencyDescriptor>();
            var typesInSpaces = target.Types;
            if (typesInSpaces.Count == 0) return 0;
            int relationships = 0;
            foreach (var typeDescriptor in typesInSpaces)
            {
                var dependencies = _model.DependenciesPerType[typeDescriptor];
                internalReferences.AddRange(dependencies.Where(dependency => dependency.Source.ImplementingType.Submodule.Equals(target)));
                var types = VerifierUtilities.AggregateDependenciesToTypes(internalReferences, true);
                relationships += types.Count();
            }
            return (relationships + 1) / (double)typesInSpaces.Count;
        }

        /// <summary>
        /// Submodules that contain multiple outgoing but few incoming dependencies are less stable because of the consequences of changes in these submodules.
        /// Submodules containing more incoming dependencies are more stable because they are more difficult to change.
        /// Instability = EfferentCoupling / (EfferentCoupling AfferentCoupling)
        /// </summary>
        /// <param name="target">a submodule</param>
        /// <returns>[0, 1]</returns>
        public double Instability(SubmoduleDescriptor target)
        {
            if (!_model.DependenciesPerSubmodule.ContainsKey(target)) return 0;
            var efferent = EfferentCoupling(target);
            var afferent = AfferentCoupling(target);
            if (efferent + afferent == 0) return 1;
            return efferent / (double)(efferent + afferent);
        }

        /// <summary>
        /// The ratio of abstract and interface types in a submodule.
        /// </summary>
        /// <param name="target">a submodule</param>
        /// <returns>[0, 1]</returns>
        public double Abstractness(SubmoduleDescriptor target)
        {
            if (!_model.DependenciesPerSubmodule.ContainsKey(target)) return 0;
            var typesInModule = target.Types;
            if (typesInModule.Count == 0) return 0;
            int abstractCount = typesInModule.Count(typeDescriptor =>
                typeDescriptor.AbstractionType == AbstractionType.Interface || typeDescriptor.AbstractionType == AbstractionType.Abstract);
            return abstractCount / (double)typesInModule.Count;
        }

        /// <summary>
        /// Describes the balance between abstractness and stability of a submodule. A submodule becomes stable when there are a lot of
        /// referrers to it. If so, then it should be abstract at the same time. A submodule that is instable should be concrete, as there
        /// are no referrers relying on contracts.
        /// Values are in range [-1, +1]. Close to -1 means the submodule is concrete but has many dependent submodules, changes to the submodule will
        /// imply lots of changes to referrers. This is called the zone of pain. On the other side, a value of one means the submodule is
        /// almost completely abstract, but noone uses it. Which is the zone of uselessness.
        /// NormalizedDistance = Abstractness + Instability - 1
        /// </summary>
        /// <param name="target">a submodule</param>
        /// <returns>[-1, +1]</returns>
        public double DistanceFromMainSequence(SubmoduleDescriptor target)
        {
            var abstractness = Abstractness(target);
            var instability = Instability(target);
            return abstractness + instability - 1;
        }

        /// <summary>
        /// Lack of Cohesion Of Methods (LCOM): The single responsibility principle states that a class should not have more than one reason to change.
        /// Such a class is said to be cohesive. A high LCOM value generally pinpoints a poorly cohesive class. 
        /// The LCOM HS (HS stands for Henderson-Sellers) takes its values in the range [0-2]. 
        /// A LCOM HS value higher than 1 should be considered alarming.
        /// </summary>
        /// <param name="target">a type</param>
        /// <returns>[0, 2]</returns>
        public double LackOfCohesion(TypeDescriptor target)
        {
            if (!_model.DependenciesPerType.ContainsKey(target)) return 0;
            var numberOfMethods = NumberOfMethods(target);
            var internalAccessors = new List<int>();
            foreach (var field in target.Members.Where(x => x.MemberType == MemberType.Field))
            {
                var afferences = _model.DependenciesPerMember[field];
                int accessors = 0;
                foreach (var dependency in afferences)
                {
                    if (!dependency.Source.ImplementingType.Equals(target)) continue;
                    accessors++;
                }
                internalAccessors.Add(accessors);
            }
            double averageAcessors = 0;
            if (internalAccessors.Count > 0)
            {
                int sum = internalAccessors.Sum();
                averageAcessors = sum / (double)internalAccessors.Count;
            }
            if (numberOfMethods == 1) numberOfMethods = 0;
            return (averageAcessors - numberOfMethods) / (1.0 - numberOfMethods);
        }

        /// <summary>
        /// The sum of the cyclomatic complexity of all members from that type.
        /// </summary>
        /// <param name="target">a type</param>
        /// <returns></returns>
        public int CyclomaticComplexity(TypeDescriptor target)
        {
            return target.Members.Where(member => member.MemberType == MemberType.Method).Sum(member => CyclomaticComplexity(member));
        }

        /// <summary>
        /// Measures the number of independent linear paths through a member. Fields return 0.
        /// </summary>
        /// <param name="target">a member</param>
        /// <returns></returns>
        public int CyclomaticComplexity(MemberDescriptor target)
        {
            if (target.MemberType != MemberType.Method) return 0;
            var instructions = _model.InstructionCache.GetInstructions(target);
            return instructions.Count(instruction => InstructionCache.CyclomaticInstructions.Contains(instruction.OpCode));
        }


        /// <summary>
        /// Count the number of fields in a type.
        /// </summary>
        /// <param name="target">a type</param>
        /// <returns></returns>
        public int NumberOfFields(TypeDescriptor target)
        {
            return target.Members.Count(x => x.MemberType == MemberType.Field);
        }

        /// <summary>
        /// Count the number of methods in a type.
        /// </summary>
        /// <param name="target">a type</param>
        /// <returns></returns>
        public int NumberOfMethods(TypeDescriptor target)
        {
            return target.Members.Count(x => x.MemberType == MemberType.Method &&
                !x.Name.StartsWith("get_") && !x.Name.StartsWith("set_"));
        }

        /// <summary>
        /// Count the number of methods in a type that are overriding a member of a base class.
        /// </summary>
        /// <param name="target">a type</param>
        /// <returns></returns>
        public int NumberOfOverriddenMethods(TypeDescriptor target)
        {
            return target.Members.Count(x => x.MemberType == MemberType.Method && x.IsOverride);
        }

        /// <summary>
        /// Count the number of types that inherit from the target type. 
        /// </summary>
        /// <param name="target"></param>
        /// <returns></returns>
        public int NumberOfChildren(TypeDescriptor target)
        {
            return target.ImplementingTypes == null ? 0 : target.ImplementingTypes.Count;
        }

        /// <summary>
        /// The deeper the inheritance graph and the more members are overridden methods, the more complex
        /// a system gets. A high value of specialization is an indicator for lack of maintainability for a
        /// type and its callers.
        /// SIX = NumberOfOverriddenMethods * DepthOfInheritance / Max(NumberOfMethods, 1)
        /// A value above 1.2 should be considered alarming.
        /// </summary>
        /// <param name="target">a type</param>
        /// <returns>[0, +)</returns>
        public double SpecializationIndex(TypeDescriptor target)
        {
            double numberOfMethods = Math.Max(1, NumberOfMethods(target));
            return NumberOfOverriddenMethods(target) * DepthOfInheritance(target) / numberOfMethods;
        }

        /// <summary>
        /// Sum of the instructions of all members
        /// </summary>
        /// <param name="target">a member</param>
        /// <returns></returns>
        public int NumberOfInstructions(TypeDescriptor target)
        {
            return target.Members.Where(member => member.MemberType == MemberType.Method).Sum(member => NumberOfInstructions(member));
        }

        /// <summary>
        /// Count the number of intermediat code instructions of a member.
        /// Fields return 0.
        /// </summary>
        /// <param name="target">a member</param>
        /// <returns></returns>
        public int NumberOfInstructions(MemberDescriptor target)
        {
            if (target.MemberType != MemberType.Method) return 0;
            var instructions = _model.InstructionCache.GetInstructions(target);
            return instructions.Count;
        }

        /// <summary>
        /// Return the depth of the inheritance tree from this type to System.Object inclusive.
        /// Interfaces do not count. Only Abstract and concrete types.
        /// The deeper a class in the hierarchy, the greater the number of methods it will probably
        /// inherit which makes it harder to predict its behavior. Deeper trees involve greater 
        /// design complexity since more classes and methods are involved. Deeper classes in the
        /// tree have a greater potential for reusing inherited methods.
        /// </summary>
        /// <param name="target"></param>
        /// <returns></returns>
        public int DepthOfInheritance(TypeDescriptor target)
        {
            int depth = 0;
            var currentType = target;
            while (currentType != null && currentType.InheritedTypes != null)
            {
                var selectedType = currentType;
                currentType = null;
                foreach (var type in selectedType.InheritedTypes)
                {
                    if (type.AbstractionType != AbstractionType.Abstract &&
                        type.AbstractionType != AbstractionType.Concrete)
                        continue;
                    currentType = type;
                    depth++;
                    break;
                }
            }
            return depth;
        }

        /// <summary>
        /// The number of members from other types that are directly accessed by the specified type.
        /// </summary>
        /// <param name="target">a type</param>
        /// <returns></returns>
        public int AssociationBetweenClasses(TypeDescriptor target)
        {
            if (!_model.DependenciesPerTypeOut.ContainsKey(target)) return 0;
            var afferences = _model.DependenciesPerTypeOut[target];
            var referredMembers = VerifierUtilities.AggregateDependenciesToMembers(afferences, false);
            return referredMembers.Count();
        }

        /// <summary>
        /// Count all interface the current type implements, which equals the number of
        /// different contracts a type has to fulfill. Interfaces on inherited types and interfaces that are
        /// base types for other interfaces also count.
        /// Low values are desired.
        /// </summary>
        /// <param name="target">a type</param>
        /// <returns></returns>
        public int NumberOfInterfaces(TypeDescriptor target)
        {
            var interfaces = new HashSet<TypeDescriptor>();
            AddInterfacesRecursive(target, interfaces);
            return interfaces.Count;
        }

        private static void AddInterfacesRecursive(TypeDescriptor target, HashSet<TypeDescriptor> interfaces)
        {
            if (target.InheritedTypes == null) return;
            foreach (var type in target.InheritedTypes)
            {
                AddInterfacesRecursive(type, interfaces);
                if (type.AbstractionType != AbstractionType.Interface) continue;
                if (interfaces.Contains(type)) continue;
                interfaces.Add(type);
            }
        }

        /// <summary>
        /// Return the deepest nesting of a members IL instructions. An implemented
        /// method at least has a nesting depth of 1 as entering the method counts as scope.
        /// Fields return 0.
        /// Anything above 5 should be considered for refactoring.
        /// </summary>
        /// <param name="target">a member</param>
        /// <returns></returns>
        public int NestingDepth(MemberDescriptor target)
        {
            if (target == null || target.MemberType != MemberType.Method) return 0;
            var instructions = _model.InstructionCache.GetInstructions(target);
            var activeBranches = new Stack<Instruction>();
            int maximumDepth = 0;
            foreach (var instruction in instructions)
            {
                while (activeBranches.Count > 0 && instruction.Offset >= (long)(activeBranches.Peek().Operand))
                {
                    activeBranches.Pop();
                }
                long lastBranch = (activeBranches.Count > 0) ? (long)(activeBranches.Peek().Operand) : 0;
                if (!InstructionCache.NestingInstructions.Contains(instruction.OpCode)) continue;
                if (lastBranch > 0 && (long)instruction.Operand >= lastBranch) continue;
                activeBranches.Push(instruction);
                maximumDepth = Math.Max(maximumDepth, activeBranches.Count);
            }
            return maximumDepth + 1;
        }

        /// <summary>
        /// Count the number of parameters of a method. For more than 2 parameters you must have good reasons.
        /// More than 3 parameters might be caused by implementing framework interfaces. Otherwise having more
        /// than 3 parameters is a hint for braking the single responsibility principle.
        /// </summary>
        /// <param name="target">a member</param>
        /// <returns></returns>
        public int NumberOfParameters(MemberDescriptor target)
        {
            return target.ParameterCount;
        }

        /// <summary>
        /// Count the number of local variables in a member. Fields return 0.
        /// </summary>
        /// <param name="target">a member</param>
        /// <returns></returns>
        public int NumberOfVariables(MemberDescriptor target)
        {
            return target.VariableCount;
        }

        /// <summary>
        /// Returns the basic of all halstead metrics for the part of program. It contains
        /// - total operators
        /// - total operands
        /// - distinct operators
        /// - distinct operands
        /// </summary>
        /// <typeparam name="T">MemberDescriptor, TypeDescriptor, SubmoduleDescriptor, ModuleDescriptor</typeparam>
        /// <param name="target">part of the program</param>
        /// <returns></returns>
        public HalsteadEntry HalsteadEntry<T>(T target) where T : IDescriptor
        {
            return _halsteadCache.GetCacheEntry(target, _model);
        }

        /// <summary>
        /// Sum of the total number of operators and operands in the part of the program.
        /// 
        /// Length = operators + operands
        /// </summary>
        /// <typeparam name="T">MemberDescriptor, TypeDescriptor, SubmoduleDescriptor, ModuleDescriptor</typeparam>
        /// <param name="target">part of the program</param>
        /// <returns></returns>
        public int HalsteadProgramLength<T>(T target) where T : IDescriptor
        {
            var cacheEntry = _halsteadCache.GetCacheEntry(target, _model);
            return cacheEntry.TotalOperators + cacheEntry.TotalOperands;
        }

        /// <summary>
        /// Count of the distinct number of operators and operands in the part of the program.
        /// 
        /// Vocabulary = ||operators|| + ||operands||
        /// </summary>
        /// <typeparam name="T">MemberDescriptor, TypeDescriptor, SubmoduleDescriptor, ModuleDescriptor</typeparam>
        /// <param name="target">part of the program</param>
        /// <returns></returns>
        public int HalsteadProgramVocabulary<T>(T target) where T : IDescriptor
        {
            var cacheEntry = _halsteadCache.GetCacheEntry(target, _model);
            return cacheEntry.DistinctOperators + cacheEntry.DistinctOperands;
        }

        /// <summary>
        /// According to Halstead, Program Volume V corresponds to the minimum number of bits required for program coding.
        /// 
        /// Volume = Length * Log2(Vocabulary)
        /// </summary>
        /// <typeparam name="T">MemberDescriptor, TypeDescriptor, SubmoduleDescriptor, ModuleDescriptor</typeparam>
        /// <param name="target"></param>
        /// <returns></returns>
        public double HalsteadVolume<T>(T target) where T : IDescriptor
        {
            return HalsteadProgramLength(target) * Math.Log(HalsteadProgramVocabulary(target), 2);
        }

        /// <summary>
        /// The difficulty level or error proneness of the program is proportional to the number of unique operators in the program.
        /// Difficulty is also proportional to the ration between the total number of operands and the number of unique operands (i.e. if the same operands are used many times in the program, it is more prone to errors). 
        ///
        /// Difficulty = ( ||operators|| / 2 ) * ( operands / ||operands|| )
        /// </summary>
        /// <typeparam name="T">MemberDescriptor, TypeDescriptor, SubmoduleDescriptor, ModuleDescriptor, ProjectDescriptor</typeparam>
        /// <param name="target">part of the program</param>
        /// <returns></returns>
        public double HalsteadDifficulty<T>(T target) where T : IDescriptor
        {
            var cacheEntry = _halsteadCache.GetCacheEntry(target, _model);
            if (cacheEntry.DistinctOperands == 0)
            {
                return 0;
            }
            return 0.5 * cacheEntry.DistinctOperators * cacheEntry.TotalOperands / cacheEntry.DistinctOperands;
        }

        /// <summary>
        /// The effort to implement or understand a program is proportional to the volume and to the difficulty level of the program. 
        ///
        /// Effort = Volume * Difficulty
        /// </summary>
        /// <typeparam name="T">MemberDescriptor, TypeDescriptor, SubmoduleDescriptor, ModuleDescriptor, ProjectDescriptor</typeparam>
        /// <param name="target">part of the program</param>
        /// <returns></returns>
        public double HalsteadEffort<T>(T target) where T : IDescriptor
        {
            return HalsteadDifficulty(target) * HalsteadVolume(target);
        }

        /// <summary>
        /// Halstead's delivered bugs is an estimate for the number of errors in the implementation.
        /// 
        /// Bugs = ( Effort ^ (2/3) ) / 3000
        ///  
        /// Delivered bugs is calculated for all modules in the current dependency model that are marked with
        /// IsPartOfProject.
        /// </summary>
        /// <returns></returns>
        public double DeliveredBugs()
        {
            return DeliveredBugs(_model.Project);
        }

        /// <summary>
        /// Halstead's delivered bugs is an estimate for the number of errors in the implementation.
        /// 
        /// Bugs = ( Effort ^ (2/3) ) / 3000
        /// </summary>
        /// <param name="project">modules that form the program</param>
        /// <returns></returns>
        public double DeliveredBugs(ProjectDescriptor project)
        {
            double effort = HalsteadEffort(project);
            return Math.Pow(effort, 2d / 3) / 3000d;
        }

        /// <summary>
        /// The time to implement or understand a program is proportional to the effort.
        /// Halstead has found that dividing the effort by 18 give an approximation for the time in seconds. 
        ///
        /// Time = Effort / 18
        /// 
        /// Time to program is calculated for all modules in the current dependency model that are marked with
        /// IsPartOfProject.
        /// </summary>
        /// <returns></returns>
        public double TimeToProgram()
        {
            return TimeToProgram(_model.Project);
        }

        /// <summary>
        /// The time to implement or understand a program (T) is proportional to the effort.
        /// Halstead has found that dividing the effort by 18 give an approximation for the time in seconds. 
        ///
        /// Time = Effort / 18
        /// </summary>
        /// <param name="project">modules that form the program</param>
        /// <returns></returns>
        public double TimeToProgram(ProjectDescriptor project)
        {
            double totalEffort = HalsteadEffort(project);
            return totalEffort / 18.0;
        }

        public double MaintainabilityIndex<T>(T target) where T : IDescriptor
        {
            //Maintainability Index = MAX(0,(171 - 5.2 * log(Halstead Volume) - 0.23 * (Cyclomatic Complexity) - 16.2 * log(Lines of Code))*100 / 171)

            return 0;
        }
    }
}
