﻿using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.IO;
using System.IO.Pipes;
using DependencyAnalysis;
using DependencyAnalysis.DataModel;
using DependencyAnalysis.Serialization;

using theeg.DependencyAnalysis.CommonAddin.Behaviors;

namespace theeg.DependencyAnalysis.CommonAddin.Models
{
    /// <summary>
    /// 
    /// </summary>
    public class AddinModel : INotifyPropertyChanged
    {
        private readonly RelayCommand _reloadCommand;
        public RelayCommand ReloadCommand { get { return _reloadCommand; } }
        private readonly RelayCommand _analyzeCommand;
        public RelayCommand AnalyzeCommand { get { return _analyzeCommand; } }
        public ISolution Solution { get; private set; }

        public bool AnalyzeInProcess { get; set; }

        public IEnumerable<IProject> Projects
        {
            get { return _projects; }
            private set
            {
                if (Equals(value, _projects)) return;
                _projects = value;
                RaisePropertyChanged("Projects");
            }
        }

        private bool _isEnabled = true;
        public bool IsEnabled
        {
            get { return _isEnabled; }
            set
            {
                if (value == _isEnabled) return;
                _isEnabled = value;
                RaisePropertyChanged("IsEnabled");
            }
        }

        public DataContractExpandoObject Metrics
        {
            get { return _metrics; }
            set
            {
                if (Equals(value, _metrics)) return;
                _metrics = value;
                RaisePropertyChanged("Metrics");
            }
        }

        public IDependencyModel DependencyModel
        {
            get { return _dependencyModel; }
            set
            {
                if (Equals(value, _dependencyModel)) return;
                _dependencyModel = value;
                _proximityModel.DependencyModel = value;
                _architectureGraphModel.DependencyModel = value;
                RaisePropertyChanged("DependencyModel");
            }
        }

        private readonly ProximityModel _proximityModel = new ProximityModel();
        public ProximityModel ProximityModel
        {
            get { return _proximityModel; }
        }

        private readonly ArchitectureGraphModel _architectureGraphModel = new ArchitectureGraphModel();
        public ArchitectureGraphModel ArchitectureGraphModel
        {
            get { return _architectureGraphModel; }
        }

        private IEnumerable<IProject> _projects;
        private DataContractExpandoObject _metrics;
        private IDependencyModel _dependencyModel;

        public AddinModel(ISolution solution)
        {
            Solution = solution;
            RefreshProjects();
            Preferences.Load(Solution, Projects);
            _reloadCommand = new RelayCommand(x =>
                {
                    RefreshProjects();
                    Preferences.Load(Solution, Projects);
                });
            _analyzeCommand = new RelayCommand(x =>
                {
                    Preferences.Store(Solution, Projects);
                    AnalyzeSolution();
                });
        }

        private void RefreshProjects()
        {
            Solution.Refresh();
            Projects = Solution.Projects;
        }

        public void AnalyzeSolution()
        {
            IsEnabled = false;
            var solution = Solution;
            solution.BuildFinished = BuildFinished;
            solution.Build();
        }

        private void BuildFinished()
        {
            var modules = new List<string>();
            foreach (var project in Projects)
            {
                if (!project.IsSelected) continue;
                var assemblyPath = project.OutputFileName;
                if (string.IsNullOrWhiteSpace(assemblyPath)) continue;
                modules.Add(assemblyPath);
            }
            StartAnalysis(modules.ToArray(), false, "DependencyAnalysisDomain_Generic");
            IsEnabled = true;
        }

        //TODO elaborate on isTest flag
        public void StartAnalysis(string[] modules, bool isTest, string pipeName)
        {
            if (AnalyzeInProcess)
            {
                var setup = Setup.BuildStandardSetup();
                setup.StartAnalysis(modules);
                DependencyModel = setup.Model;
                var metricsReport = new MetricsReport { Setup = setup };
                Metrics = (DataContractExpandoObject)metricsReport.FetchReport<DataContractExpandoObject>();
                //Principles.EvaluateMethodenreinheit(DependencyModel);
            }
            else
            {
                var analysisHost = new Process
                {
                    StartInfo = new ProcessStartInfo(typeof(AnalysisHost.App).Assembly.Location)
                    {
                        UseShellExecute = false,
                        Arguments = pipeName
                    }
                };
                using (var pipeServer = new NamedPipeServerStream(pipeName, PipeDirection.InOut))
                {
                    if (!isTest)
                    {
                        analysisHost.Start();
                    }
                    pipeServer.WaitForConnection();
                    try
                    {
                        var writer = new StreamWriter(pipeServer);
                        var reader = new StreamReader(pipeServer);

                        writer.AutoFlush = true;
                        writer.WriteLine(AnalysisHost.App.Synchronization);
                        pipeServer.WaitForPipeDrain();
                        foreach (var module in modules)
                        {
                            writer.WriteLine(module);
                        }
                        pipeServer.WaitForPipeDrain();
                        writer.WriteLine(AnalysisHost.App.Synchronization);
                        pipeServer.WaitForPipeDrain();

                        AnalysisHost.App.WaitForSynchronization(reader, () => !isTest && analysisHost.HasExited);
                        var result = (AnalysisHost.AnalysisResult)AnalysisHost.App.DeserializeFromStream(pipeServer);
                        DependencyModel = result.DependencyModel;
                        Metrics = result.Metrics;
                    }
                    catch (IOException)
                    {
                    }
                }
                if (!isTest)
                {
                    analysisHost.WaitForExit();
                }
            }
        }

        public void SelectCommandExecuted(object parameter)
        {

        }

        public event PropertyChangedEventHandler PropertyChanged;

        protected virtual void RaisePropertyChanged(string propertyName)
        {
            var handler = PropertyChanged;
            if (handler != null) handler(this, new PropertyChangedEventArgs(propertyName));
        }
    }
}
